from quantplay.broker import motilal
import hashlib
# import requests
# import json
# import pyotp
#
# class MotilalTest:
#     headers = {
#         "Accept": "application/json",
#         "ApiKey": "ypdJRIbZkCTHV5Tp",
#         "vendorinfo": "EMUM755714",
#         "User-Agent": "MOSL/V.1.1.0",
#         "SourceId": "WEB",
#         "MacAddress": "00:50:56:BD:F4:0B",
#         "ClientLocalIp": "192.168.165.165",
#         "ClientPublicIp": "106.193.137.95",
#         "osname": "Ubuntu",
#         "osversion": "10.0.19041",
#         "devicemodel": "AHV",
#         "manufacturer": "DELL",
#         "productname": "Your Product Name",
#         "productversion": "Your Product Version",
#         "installedappid": "AppID",
#         "browsername": "Chrome",
#         "browserversion": "105.0"
#     }
#     url = "https://uatopenapi.motilaloswal.com/rest/login/v3/authdirectapi"
#     ltp_utl = "https://uatopenapi.motilaloswal.com/rest/report/v1/getltpdata"
#     place_order_url = "https://uatopenapi.motilaloswal.com/rest/trans/v1/placeorder"
#
#     def generate_token(self, totp):
#         str = "Quant@123ypdJRIbZkCTHV5Tp"
#         result = hashlib.sha256(str.encode())
#
#         data = {
#             "userid": "EMUM755714",
#             "password": result.hexdigest(),
#             "2FA": "25/03/1993",
#             "totp": totp
#         }
#
#         response = requests.post(MotilalTest.url, headers=MotilalTest.headers, data=json.dumps(data))
#         resp_json = response.json()
#         print(resp_json)
#
#         auth_token = resp_json['AuthToken']
#
#         MotilalTest.headers['Authorization'] = auth_token
#
#     def get_ltp(self):
#         data = {
#             "userid": "EMUM755714",
#             "exchange": "NSE",
#             "scripcode": 3045
#         }
#
#         response = requests.post(MotilalTest.ltp_utl, headers=MotilalTest.headers, data=json.dumps(data))
#         print(response.json())
#         return response.json()["data"]["ltp"] / 100.0
#
#     def place_order(self, price, quantity):
#         data = {
#             "exchange": "NSE",
#             "symboltoken": 3045,
#             "buyorsell": "BUY",
#             "ordertype": "LIMIT",
#             "producttype": "NORMAL",
#             "orderduration": "DAY",
#             "price": price,
#             "triggerprice": 0,
#             "quantityinlot": quantity,
#             "disclosedquantity": 0,
#             "amoorder": "N",
#             "algoid": "",
#             "tag": "test"
#         }
#
#         response = requests.post(MotilalTest.place_order_url, headers=MotilalTest.headers, data=json.dumps(data)).json()
#         print(response)
#
#
# motilal = MotilalTest()
# secret_key = "7TGIDZBCUCCF66Z3ANJIEZBVALQUB7MN"
# totp = pyotp.TOTP(secret_key)
# current_totp = totp.now()
# motilal.generate_token(current_totp)
# # motilal.send_otp()
# # motilal.verify_otp(otp)
# # ltp = motilal.get_ltp("NSE", "SBIN")
# ltp = motilal.get_ltp()
# print("LTP: %d", ltp)
#
# motilal.place_order(ltp, 1)
#
# # motilal.place_order("SBIN", "NSE", 1, "LIMIT", "BUY", "test", "NORMAL", ltp, 0)
# # orders_placed = motilal.get_orders()
# # print(orders_placed)
# #
# # motilal.modify_orders_till_complete(orders_placed)

motilal = Motilal(is_uat=True)
motilal.generate_token()


from quantplay.services.market import Market
from quantplay.services.tradelens import TradeLens
from quantplay.backtest.backtest_trades import Backtesting

market = Market()
backtesting = Backtesting(market)
tradelens = TradeLens




import numpy as np

class TradeLens:

    @staticmethod
    def max_drawdowns(trades):
        trades = trades.sort_values("order_timestamp")
        temp_df = trades[['order_timestamp', 'profit']]

        for j in range(0, 5):
            temp_df.loc[:, 'balance'] = temp_df.profit.cumsum()
            temp_df.loc[:, 'running_max_balance'] = np.maximum.accumulate(temp_df.balance)
            temp_df.loc[:, 'drawdowns'] = temp_df.running_max_balance - temp_df.balance

            max_drawdown = np.max(temp_df.drawdowns)
            end_index = None
            for i in range(0, len(temp_df)):
                end_index = len(temp_df) - 1 - i
                if temp_df.iloc[end_index]['drawdowns'] != max_drawdown:
                    continue

                start_index = end_index
                while temp_df.iloc[start_index - 1]['running_max_balance'] == temp_df.iloc[start_index]['running_max_balance']:
                    start_index = start_index - 1

                print("Max drawdown {} from {} till {} amount {}".format(j,
                                                               temp_df.iloc[start_index]['order_timestamp'],
                                                               temp_df.iloc[end_index]['order_timestamp'],
                                                                        max_drawdown))

                temp_df = temp_df.drop(temp_df.index[start_index:end_index])
                break


    @staticmethod
    def analyse(trades):
        exchanges = list(trades.exchange.unique())
        trades.loc[:, 'segment'] = np.where("PE" == trades.tradingsymbol.str[-2:], "PE", "CE")

        trades.loc[:, 'hour'] = trades.date.dt.hour
        trades.loc[:, "day_of_week"] = trades.date.dt.day_name()

        print(trades.groupby('hour').profit.mean())

        print(trades.groupby('day_of_week').profit.mean())

        print(trades.groupby(['hour', 'day_of_week']).profit.mean())

        if len(exchanges) == 1 and exchanges[0] == "NFO":
            print(trades.groupby('segment').profit.mean())
            print(trades.groupby(['segment', 'day_of_week']).profit.mean())
            print(trades.groupby(['hour', 'segment']).profit.mean())

        trades.loc[:, 'trade_return'] = trades.profit / trades.exposure
        trades.loc[:, 'time'] = trades.order_timestamp.dt.time.astype(str)
        print("Trade return by time")
        print(trades.groupby('time').trade_return.mean())

        trades.loc[:, 'trade_return'] = trades.close_price/trades.entry_price - 1
        print("Mean return of trades")
        print(trades.trade_return.mean())

        trades.loc[:, 'week_number'] = (trades.order_timestamp.dt.day/7).astype(int)
        print("profit by week number")
        print(trades.groupby('week_number').profit.mean())




if __name__ == "__main__":
    print("testing")

    tradingsymbol = "SBIN"
    transaction_type = "BUY"
    exchange = "NSE"
    quantity = 1
    order_type = "LIMIT"
    price = 550
    product = "MIS"
    tag = "test"


import time
import traceback

from retrying import retry
import getpass
from quantplay.utils.selenium_utils import Selenium
from quantplay.config.qplay_config import QplayConfig
import pyotp

class KiteUtils():
    zerodha_username = "zerodha_username"
    zerodha_password = "zerodha_password"
    zeordha_totp_unique_id = "zerodha_totp_unique_id"

    @staticmethod
    @retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_request_token(api_key=None, user_id=None, password=None, totp=None):
        try:
            browser = Selenium.get_browser()

            # TODO api should be fetched from configuration

            kite_url = "https://kite.trade/connect/login?api_key={}&v=3".format(api_key)
            print("Kite Url {}".format(kite_url))
            browser.get(kite_url)
            time.sleep(5)

            user_id_element = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[1]/input')
            password_element = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[2]/input')

            user_id_element.send_keys(user_id)
            password_element.send_keys(password)

            login_attempt = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[4]/button')
            login_attempt.submit()
            time.sleep(5)

            kite_pin = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[2]/input')
            kite_pin.send_keys(pyotp.TOTP(totp).now())
            time.sleep(3)
            # login_attempt = browser.find_element("xpath", '//*[@id="container"]/div/div/div[2]/form/div[3]/button')
            # login_attempt.submit()
            # time.sleep(1)

            url = browser.current_url
            print("got kite url {}".format(url))
            request_token = url.split('token=')[1].split('&')[0]

            browser.close()

            return request_token
        except:
            print(traceback.print_exc())
            raise





import copy
import math
import random
import time
from collections import defaultdict
from datetime import datetime
from datetime import timedelta
from threading import Lock

import numpy as np

from quantplay.exception.exceptions import QuantplayOrderPlacementException
from quantplay.utils.constant import Constants
from quantplay.utils.exchange import Market as MarketConstants
from quantplay.utils.number_utils import NumberUtils

logger = Constants.logger

class Broker():

    def __init__(self):
        self.instrument_id_to_symbol_map = dict()
        self.instrument_id_to_exchange_map = dict()
        self.instrument_id_to_security_type_map = dict()
        self.exchange_symbol_to_instrument_id_map = defaultdict(dict)
        self.order_type_sl = "SL"
        self.nfo_exchange = "NFO"
        self.lock = Lock()

    def round_to_tick(self, number):
        return round(number * 20) / 20

    def populate_instruments(self, instruments):
        """Fetches instruments for all exchanges from the broker
        and stores them in the member attributes.
        """
        Constants.logger.info("populating instruments")
        for instrument in instruments:
            exchange, symbol, instrument_id = (
                instrument.exchange,
                instrument.symbol,
                instrument.instrument_id,
            )
            self.instrument_id_to_symbol_map[instrument_id] = symbol
            self.instrument_id_to_exchange_map[instrument_id] = exchange
            self.instrument_id_to_security_type_map[
                instrument_id
            ] = instrument.security_type()
            self.exchange_symbol_to_instrument_id_map[exchange][symbol] = instrument_id

    def execute_order_v2(self, order):
        start_time = datetime.now()
        tradingsymbol = order['tradingsymbol']
        exchange = order['exchange']
        trigger_price = order['trigger_price']
        transaction_type = order['transaction_type']
        if order['validity'] is not None and order['trigger_price'] is not None:
            while True:
                self.lock.acquire()
                try:
                    ltp = self.get_ltp(exchange, tradingsymbol)
                except Exception as e:
                    Constants.logger.error("[GET_LTP_FAILED] with exception {}".format(e))
                time.sleep(.5)
                self.lock.release()
                if (transaction_type == "SELL" and trigger_price > ltp) or (
                        transaction_type == "BUY" and trigger_price < ltp):
                    logger.info("[EXECUTING_ORDER] ltp {} crossed trigger price {} for {}".format(ltp, trigger_price, order))
                    self.execute_order(tradingsymbol=order['tradingsymbol'],
                                       exchange=order['exchange'],
                                       quantity=order['quantity'],
                                       product=order['product'],
                                       tag=order['tag'],
                                       stoploss=order['stoploss'],
                                       transaction_type=order['transaction_type'],
                                       order_type=order['order_type'])
                    return
                current_time = datetime.now()
                if (current_time - start_time).seconds > order['validity']:
                    Constants.logger.info("[ORDER_VALIDITY_EXPIRED] order [{}]".format(order))
                    return

    def execute_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                      stoploss=None, tag=None, product=None, price=None):
        if price is None:
            price = self.get_ltp(exchange=exchange, tradingsymbol=tradingsymbol)
            trade_price = copy.deepcopy(price)
        try:
            if stoploss != None:
                if transaction_type == "SELL":
                    sl_transaction_type = "BUY"
                    sl_trigger_price = self.round_to_tick(price * (1 + stoploss))

                    if exchange == self.nfo_exchange:
                        price = sl_trigger_price * 1.05
                    elif exchange == "NSE":
                        price = sl_trigger_price * 1.01
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                elif transaction_type == "BUY":
                    sl_transaction_type = "SELL"
                    sl_trigger_price = self.round_to_tick(price * (1 - stoploss))

                    if exchange == self.nfo_exchange:
                        price = sl_trigger_price * .95
                    elif exchange == "NSE":
                        price = sl_trigger_price * .99
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                else:
                    raise Exception("Invalid transaction_type {}".format(transaction_type))
                stoploss_order_id = self.place_order(tradingsymbol=tradingsymbol,
                                                     exchange=exchange,
                                                     quantity=quantity,
                                                     order_type=self.order_type_sl,
                                                     transaction_type=sl_transaction_type,
                                                     tag=tag, product=product, price=sl_price,
                                                     trigger_price=sl_trigger_price)

                if stoploss_order_id is None:
                    Constants.logger.error(
                        "[ORDER_REJECTED] tradingsymbol {}".format(tradingsymbol))
                    raise QuantplayOrderPlacementException("Order reject for {}".format(tradingsymbol))

            if order_type == "MARKET":
                trade_price = 0

            response = self.place_order(tradingsymbol=tradingsymbol, exchange=exchange, quantity=quantity,
                                        order_type=order_type, transaction_type=transaction_type, tag=tag,
                                        product=product, price=trade_price)
            return response
        except Exception as e:
            raise e

    """
            Input  : quantplay symbol
            Output : broker symbol
        """

    def get_symbol(self, symbol):
        return symbol

    """
        Input  : quantplay exchange
        Output : broker exchange
    """

    def get_exchange(self, exchange):
        return exchange

    def place_order_quantity(self, quantity, tradingsymbol, exchange):
        lot_size = self.get_lot_size(exchange, tradingsymbol)
        quantity_in_lots = int(quantity/lot_size)

        return quantity_in_lots*lot_size

    def get_product(self, product):
        return product

    def get_lot_size(self, exchange, tradingsymbol):
        try:
            return int(self.symbol_data["{}:{}".format(exchange, tradingsymbol)]['lot_size'])
        except Exception as e:
            logger.error("[GET_LOT_SIZE] unable to get lot size for {} {}".format(exchange, tradingsymbol))
            raise e

    def option_symbol(self, underlying_symbol, expiry_date, strike_price, type):
        option_symbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[underlying_symbol]
        option_symbol += expiry_date.strftime('%y')

        month_number = str(int(expiry_date.strftime("%m")))
        monthly_option_prefix = expiry_date.strftime("%b").upper()

        if int(month_number) >= 10:
            week_option_prefix = monthly_option_prefix[0]
        else:
            week_option_prefix = month_number
        week_option_prefix += expiry_date.strftime("%d")

        next_expiry = expiry_date + timedelta(days=7)

        if next_expiry.month != expiry_date.month:
            option_symbol += monthly_option_prefix
        else:
            option_symbol += week_option_prefix

        option_symbol += str(int(strike_price))
        option_symbol += type

        return option_symbol

    def square_off_by_tag(self, tag, dry_run=True, sleep_time=0.05):
        orders = self.orders(tag=tag)
        orders.loc[:, 'exit_quantity'] = np.where(orders.transaction_type == "BUY",
                                                  -orders.quantity_traded,
                                                  orders.quantity_traded)
        exit_orders = orders.groupby('tradingsymbol').agg({'exit_quantity' : 'sum',
                                                           'exchange' : 'first',
                                                           'product' : 'first'
                                                           }).reset_index()

        orders_to_close = []
        exit_orders = exit_orders[exit_orders.exit_quantity != 0]
        positions = exit_orders.to_dict('records')
        for position in positions:
            exchange = position['exchange']
            tradingsymbol = position['tradingsymbol']
            quantity = position['exit_quantity']

            transaction_type = "SELL"
            if quantity == 0:
                continue
            elif quantity > 0:
                transaction_type = "BUY"

            quantity = abs(quantity)
            quantity_in_lots = int(quantity / self.get_lot_size(exchange, tradingsymbol))

            split_into = int(math.ceil(quantity_in_lots / 25))
            split_array = NumberUtils.split(abs(quantity_in_lots), abs(split_into))

            for q in split_array:
                orders_to_close.append(
                    {
                        'tradingsymbol': tradingsymbol,
                        'exchange': exchange,
                        'transaction_type': transaction_type,
                        'quantity_in_lots': q,
                        'product' : position['product']
                    })

        random.shuffle(orders_to_close)
        orders_to_close = sorted(orders_to_close, key=lambda d: d['transaction_type'])
        for order in orders_to_close:
            tradingsymbol = order['tradingsymbol']
            exchange = order['exchange']
            transaction_type = order['transaction_type']
            product = order['product']
            quantity = order['quantity_in_lots']*self.get_lot_size(exchange, tradingsymbol)
            quantity = self.place_order_quantity(quantity, tradingsymbol, exchange)

            print(tradingsymbol, exchange, transaction_type, quantity)
            if dry_run == False:
                self.place_order(tradingsymbol=tradingsymbol,
                                 exchange=exchange,
                                 quantity=quantity,
                                 order_type="MARKET",
                                 transaction_type=transaction_type,
                                 tag=tag,
                                 product=product,
                                 price=0)
                time.sleep(sleep_time)

        return orders_to_close


from quantplay.broker.finvasia_utils.shoonya import ShoonyaApiPy
from quantplay.utils.constant import Constants, OrderType
from quantplay.broker.generics.broker import Broker
from quantplay.exception.exceptions import InvalidArgumentException
import pyotp
import io
import os
import zipfile

import pandas as pd
import numpy as np
import requests
from retrying import retry
import time
import json
import traceback
import copy

logger = Constants.logger


class FinvAsia(Broker):

    def __init__(self,
                 order_updates=None,
                 api_secret=None,
                 imei=None,
                 password=None,
                 totp_key=None,
                 user_id=None,
                 vendor_code=None,
                 load_instruments=True):
        super(FinvAsia, self).__init__()
        self.order_updates = order_updates

        self.api = ShoonyaApiPy()
        totp = pyotp.TOTP(totp_key).now()
        response = self.api.login(userid=user_id,
                                  password=password,
                                  twoFA=totp,
                                  vendor_code=vendor_code,
                                  api_secret=api_secret,
                                  imei=imei)
        print("finvasia login successful email {} account_id {}".format(response['email'], response['actid']))

        self.set_attributes(response)

        if load_instruments:
            self.load_instrument()

        self.order_type_sl = "SL-LMT"

    def set_attributes(self, response):
        self.email = response['email']
        self.user_id = response['actid']
        self.full_name = response['uname']

    def initialize_symbol_data(self):
        instruments = self.instrument_data
        instruments = instruments.to_dict('records')
        self.symbol_data = {}
        for instrument in instruments:
            exchange = instrument['Exchange']
            tradingsymbol = instrument['TradingSymbol']

            instrument_data = copy.deepcopy(instrument)
            instrument_data['lot_size'] = instrument['LotSize']
            self.symbol_data["{}:{}".format(exchange, tradingsymbol)] = instrument_data

    def initialize_expiry_fields(self):
        self.instrument_data.loc[:, 'tradingsymbol'] = self.instrument_data.Symbol
        self.instrument_data.loc[:, 'expiry'] = pd.to_datetime(self.instrument_data.Expiry)

        self.instrument_data.loc[:, "expiry_year"] = self.instrument_data["expiry"].dt.strftime("%y").astype(str)
        self.instrument_data.loc[:, "month"] = self.instrument_data["expiry"].dt.strftime("%b").str.upper()

        self.instrument_data.loc[:, "month_number"] = self.instrument_data["expiry"].dt.strftime("%m").astype(
            float).astype(str)
        self.instrument_data.loc[:, 'month_number'] = np.where(self.instrument_data.month_number == 'nan',
                                                               np.nan,
                                                               self.instrument_data.month_number.str.split(
                                                                   ".").str[0]
                                                               )

        self.instrument_data.loc[:, "week_option_prefix"] = np.where(
            self.instrument_data.month_number.astype(float) >= 10,
            self.instrument_data.month.str[0] + self.instrument_data["expiry"].dt.strftime("%d").astype(str),
            self.instrument_data.month_number + self.instrument_data["expiry"].dt.strftime("%d").astype(str),
        )

        self.instrument_data.loc[:, "next_expiry"] = self.instrument_data.expiry + pd.DateOffset(days=7)

    def add_quantplay_fut_tradingsymbol(self):
        seg_condition = [
            ((self.instrument_data["Instrument"].str.contains("FUT")) & (
                    self.instrument_data.Instrument != "OPTFUT"))
        ]

        tradingsymbol = [
            self.instrument_data.tradingsymbol + self.instrument_data.expiry_year + self.instrument_data.month + "FUT"
        ]

        self.instrument_data.loc[:, "tradingsymbol"] = np.select(
            seg_condition, tradingsymbol, default=self.instrument_data.tradingsymbol
        )

    def add_quantplay_opt_tradingsymbol(self):
        seg_condition = (self.instrument_data["StrikePrice"] > 0)
        weekly_option_condition = (
                (self.instrument_data.expiry.dt.month == self.instrument_data.next_expiry.dt.month) & (
                self.instrument_data.Exchange == "NFO"))
        month_option_condition = (
                (self.instrument_data.expiry.dt.month != self.instrument_data.next_expiry.dt.month) | (
                self.instrument_data.Exchange == "MCX"))

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.expiry_year,
            self.instrument_data.tradingsymbol
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition & weekly_option_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.week_option_prefix,
            self.instrument_data.tradingsymbol
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition & month_option_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.month,
            self.instrument_data.tradingsymbol
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol +
            self.instrument_data.StrikePrice.astype(float).astype(str).str.split(".").str[0],
            self.instrument_data.tradingsymbol
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.OptionType,
            self.instrument_data.tradingsymbol
        )

    def get_df_from_zip(self, url):
        response = requests.get(url, timeout=10)
        z = zipfile.ZipFile(io.BytesIO(response.content))

        directory = '/tmp/'
        z.extractall(path=directory)
        file_name = url.split(".txt")[0].split("/")[-1]
        os.system('cp /tmp/{}.txt /tmp/{}.csv'.format(file_name, file_name))
        time.sleep(2)
        return pd.read_csv('/tmp/{}.csv'.format(file_name))

    def load_instrument(self):
        instrument_file_EQ = self.get_df_from_zip("https://api.shoonya.com/NSE_symbols.txt.zip")
        instrument_file_FO = self.get_df_from_zip("https://api.shoonya.com/NFO_symbols.txt.zip")
        instrument_file_MCX = self.get_df_from_zip("https://api.shoonya.com/MCX_symbols.txt.zip")

        self.instrument_data = pd.concat([instrument_file_MCX, instrument_file_FO, instrument_file_EQ])

        self.initialize_expiry_fields()
        self.add_quantplay_opt_tradingsymbol()
        self.add_quantplay_fut_tradingsymbol()
        self.fno_symbol_map = dict(zip(self.instrument_data.TradingSymbol, self.instrument_data.tradingsymbol))
        self.quantplay_symbol_map = {v: k for k, v in self.fno_symbol_map.items()}

        self.initialize_symbol_data()

    def get_symbol(self, symbol):
        if symbol not in self.quantplay_symbol_map:
            return symbol
        return self.quantplay_symbol_map[symbol]

    def get_transaction_type(self, transaction_type):
        if transaction_type == "BUY":
            return "B"
        elif transaction_type == "SELL":
            return "S"

        raise InvalidArgumentException("transaction type {} not supported for trading".format(transaction_type))

    def get_order_type(self, order_type):
        if order_type == OrderType.market:
            return "MKT"
        elif order_type == OrderType.sl:
            return "SL-LMT"
        elif order_type == OrderType.slm:
            return "SL-MKT"
        elif order_type == OrderType.limit:
            return "LMT"
        elif order_type in [self.order_type_sl]:
            return order_type

        raise InvalidArgumentException("order type {} not supported for trading".format(order_type))

    def get_product(self, product):
        if product == "NRML":
            return "M"
        elif product == "CNC":
            return "C"
        elif product == "MIS":
            return "I"
        elif product in ["M", "C", "I"]:
            return product

        raise InvalidArgumentException("Product {} not supported for trading".format(product))

    def event_handler_order_update(self, order):
        try:
            order['placed_by'] = order['actid']
            order['tag'] = order['actid']
            order['order_id'] = order['norenordno']
            order['exchange_order_id'] = order['exchordid']
            order['exchange'] = order['exch']

            # TODO translate symbol
            # -EQ should be removed
            # F&O symbol translation
            order['tradingsymbol'] = order['tsym']

            if order['exchange'] == "NSE":
                order['tradingsymbol'] = order['tradingsymbol'].replace("-EQ", "")
            elif order['exchange'] in ["NFO", "MCX"]:
                order["tradingsymbol"] = self.fno_symbol_map[order["tradingsymbol"]]

            order['order_type'] = order['prctyp']
            if order['order_type'] == "LMT":
                order['order_type'] = "LIMIT"
            elif order['order_type'] == "MKT":
                order['order_type'] = "MARKET"
            elif order['order_type'] == "SL-LMT":
                order['order_type'] = "SL"

            if order['pcode'] == "M":
                order['product'] = "NRML"
            elif order['pcode'] == "C":
                order['product'] = "CNC"
            elif order['pcode'] == "I":
                order['product'] = "MIS"

            if order['trantype'] == "S":
                order['transaction_type'] = "SELL"
            elif order['trantype'] == "B":
                order['transaction_type'] = "BUY"
            else:
                logger.error("[UNKNOW_VALUE] finvasia transaction type {} not supported".format(order['trantype']))

            order['quantity'] = int(order['qty'])

            if 'trgprc' in order:
                order['trigger_price'] = float(order['trgprc'])
            else:
                order['trigger_price'] = None

            order['price'] = float(order['prc'])

            if order["status"] == "TRIGGER_PENDING":
                order["status"] = "TRIGGER PENDING"
            elif order["status"] == "CANCELED":
                order["status"] = "CANCELLED"

            print(f"order feed {order}")
            self.order_updates.put(order)
        except Exception as e:
            logger.error("[ORDER_UPDATE_PROCESSING_FAILED] {}".format(e))

    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        try:
            if trigger_price == 0:
                trigger_price = None

            transaction_type = self.get_transaction_type(transaction_type)
            order_type = self.get_order_type(order_type)
            product = self.get_product(product)
            tradingsymbol = self.get_symbol(tradingsymbol)

            data = {
                "product_type": product,
                "buy_or_sell": transaction_type,
                "exchange": exchange,
                "tradingsymbol": tradingsymbol,
                "quantity": quantity,
                "price_type": order_type,
                "price": price,
                "trigger_price": trigger_price,
                "remarks": tag
            }
            Constants.logger.info("[PLACING_ORDER] {}".format(json.dumps(data)))
            response = self.api.place_order(buy_or_sell=transaction_type,
                                            product_type=product,
                                            exchange=exchange,
                                            tradingsymbol=tradingsymbol,
                                            quantity=quantity,
                                            discloseqty=0,
                                            price_type=order_type,
                                            price=price,
                                            trigger_price=trigger_price,
                                            retention='DAY',
                                            remarks=tag)
            Constants.logger.info("[PLACE_ORDER_RESPONSE] {} input {}".format(response, json.dumps(data)))
            if 'norenordno' in response:
                return response['norenordno']
            else:
                raise Exception(response)
        except Exception as e:
            print(traceback.print_exc())
            exception_message = "Order placement failed with error [{}]".format(str(e))
            print(exception_message)

    def get_orders(self):
        return self.api.get_order_book()

    def get_ltp(self, exchange, tradingsymbol):
        tradingsymbol = self.get_symbol(tradingsymbol)
        token = self.symbol_data["{}:{}".format(exchange, tradingsymbol)]['Token']
        return float(self.api.get_quotes(exchange, str(token))['lp'])

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def modify_order(self, data):
        try:
            logger.info("Modifying order [{}] new price [{}]".format(data['norenordno'], data['prc']))
            response = self.api.modify_order(orderno=data['norenordno'],
                                             exchange=data['exch'],
                                             tradingsymbol=data['tsym'],
                                             newprice_type=data['prctyp'],
                                             newquantity=data['qty'],
                                             newprice=data['prc'])
            logger.info("[MODIFY_ORDER_RESPONSE] order id [{}] response [{}]".format(data['norenordno'], response))
            return response
        except Exception as e:
            exception_message = "OrderModificationFailed for {} failed with exception {}".format(data['norenordno'], e)
            Constants.logger.error("{}".format(exception_message))

    def modify_price(self, order_id, price, trigger_price=None, order_type=None):
        data = {}
        order_history = self.api.single_order_history(order_id)
        order_details = order_history[0]
        data['norenordno'] = order_id
        data['prc'] = price
        data['prctyp'] = order_details['prctyp']
        data['exch'] = order_details['exch']
        data['qty'] = order_details['qty']
        data['tsym'] = order_details['tsym']
        if trigger_price != None and trigger_price > 0:
            data['newtrigger_price'] = trigger_price
        else:
            data['newtrigger_price'] = None

        self.modify_order(data)

    def cancel_order(self, order_id):
        self.api.cancel_order(order_id)

    def modify_orders_till_complete(self, orders_placed, sleep_time=10):
        modification_count = {}
        while 1:
            time.sleep(sleep_time)
            orders = pd.DataFrame(self.get_orders())

            orders = orders[orders.norenordno.isin(orders_placed)]
            orders = orders[~orders.status.isin(["REJECTED", "CANCELED", "COMPLETE"])]

            if len(orders) == 0:
                Constants.logger.info("ALL orders have been completed")
                break

            orders = orders.to_dict('records')
            for order in orders:
                order_id = order['order_id']

                ltp = self.get_ltp(order['exchange'], order['tradingsymbol'])
                order['prc'] = ltp
                self.modify_order(order)

                if order_id not in modification_count:
                    modification_count[order_id] = 1
                else:
                    modification_count[order_id] += 1

                time.sleep(.1)

                if modification_count[order_id] > 5:
                    order['prctyp'] = "MKT"
                    order['prc'] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)

    def stream_order_data(self):
        self.api.start_websocket(order_update_callback=self.event_handler_order_update)

    def profile(self):

        response = {
            'user_id': self.user_id,
            'full_name': self.full_name,
            'email' : self.email
        }

        return response


import hashlib
import requests
import json
import time
import pandas as pd

from quantplay.utils.constant import Constants, timeit
from quantplay.config.qplay_config import QplayConfig
import getpass
from quantplay.broker.generics.broker import Broker
import traceback
from quantplay.exception.exceptions import InvalidArgumentException
import numpy as np
import pyotp
import math
from quantplay.utils.number_utils import NumberUtils
import random
import websocket
import _thread as thread
import copy
import threading


class Motilal(Broker):
    user_id = "motilal_user_id"
    api_key = "motilal_api_key"
    password = "motilal_password"
    auth_token = "motilal_auth_token"
    two_factor_authentication = "motilal_2FA"
    secret_key = "motilal_secret_key"

    headers = {
        "Accept": "application/json",
        "User-Agent": "MOSL/V.1.1.0",
        "SourceId": "WEB",
        "MacAddress": "00:50:56:BD:F4:0B",
        "ClientLocalIp": "192.168.165.165",
        "ClientPublicIp": "106.193.137.95",
        "osname": "Ubuntu",
        "osversion": "10.0.19041",
        "devicemodel": "AHV",
        "manufacturer": "DELL",
        "productname": "Your Product Name",
        "productversion": "Your Product Version",
        "installedappid": "AppID",
        "browsername": "Chrome",
        "browserversion": "105.0"
    }

    @timeit(MetricName="Motilal:__init__")
    def __init__(self, is_uat=False, headers=None, load_instruments=True, user_id=None, password=None, api_key=None,
                 two_fa=None, totp=None):
        super(Motilal, self).__init__()

        if headers:
            self.headers = headers
        self.symbol_scripcode_map = {"NSE": {}, "NSEFO": {}}
        self.instrument_data_by_exchange = {}

        uat = ""
        if is_uat:
            uat = "uat"

        self.url = "https://{}openapi.motilaloswal.com/rest/login/v3/authdirectapi".format(uat)
        self.otp_url = "https://{}openapi.motilaloswal.com/rest/login/v3/resendotp".format(uat)
        self.verify_otp_url = "https://{}openapi.motilaloswal.com/rest/login/v3/verifyotp".format(uat)
        self.ltp_utl = "https://{}openapi.motilaloswal.com/rest/report/v1/getltpdata".format(uat)
        self.place_order_url = "https://{}openapi.motilaloswal.com/rest/trans/v1/placeorder".format(uat)
        self.get_profile_url = "https://{}openapi.motilaloswal.com/rest/login/v1/getprofile".format(uat)
        self.margin_summary_url = "https://{}openapi.motilaloswal.com/rest/report/v1/getreportmarginsummary".format(uat)
        self.modify_order_url = "https://{}openapi.motilaloswal.com/rest/trans/v2/modifyorder".format(uat)
        self.order_book_url = "https://{}openapi.motilaloswal.com/rest/book/v1/getorderbook".format(uat)
        self.cancel_order_url = "https://{}openapi.motilaloswal.com/rest/trans/v1/cancelorder".format(uat)
        self.positions_url = "https://{}openapi.motilaloswal.com/rest/book/v1/getposition".format(uat)

        try:
            if headers:
                self.get_orders()
            else:
                self.generate_token(user_id, password, api_key, two_fa, totp)
                self.get_orders()
        except Exception as e:
            print(traceback.print_exc())
            raise e

        if load_instruments:
            self.load_instrument()

        self.order_type_sl = "STOPLOSS"
        self.nfo_exchange = "NSEFO"
        self.exchange_code_map = {
            "NFO": "NSEFO",
            "CDS": "NSECD"
        }

    def update_headers(self):
        Constants.logger.info("Updating headers")
        quantplay_config = QplayConfig.get_config()

        auth_token = quantplay_config['DEFAULT'][Motilal.auth_token]
        api_key = quantplay_config['DEFAULT'][Motilal.api_key]
        user_id = quantplay_config['DEFAULT'][Motilal.user_id]

        self.headers['vendorinfo'] = user_id
        self.headers['Authorization'] = auth_token
        self.headers['ApiKey'] = api_key

        self.user_id = user_id

    @timeit(MetricName="Motilal:initialize_expiry_fields")
    def initialize_expiry_fields(self):
        self.instrument_file_FO.loc[:, 'tradingsymbol'] = self.instrument_file_FO.scripshortname
        self.instrument_file_FO.loc[:, 'expiry'] = pd.to_datetime(self.instrument_file_FO.expirydate + 315513000,
                                                                  unit='s')

        self.instrument_file_FO.loc[:, "expiry_year"] = self.instrument_file_FO["expiry"].dt.strftime("%y").astype(str)
        self.instrument_file_FO.loc[:, "month"] = self.instrument_file_FO["expiry"].dt.strftime("%b").str.upper()

        self.instrument_file_FO.loc[:, "month_number"] = self.instrument_file_FO["expiry"].dt.strftime("%m").astype(
            float).astype(str)
        self.instrument_file_FO.loc[:, 'month_number'] = np.where(self.instrument_file_FO.month_number == 'nan',
                                                                  np.nan,
                                                                  self.instrument_file_FO.month_number.str.split(
                                                                      ".").str[0]
                                                                  )

        self.instrument_file_FO.loc[:, "week_option_prefix"] = np.where(
            self.instrument_file_FO.month_number.astype(float) >= 10,
            self.instrument_file_FO.month.str[0] + self.instrument_file_FO["expiry"].dt.strftime("%d").astype(str),
            self.instrument_file_FO.month_number + self.instrument_file_FO["expiry"].dt.strftime("%d").astype(str),
        )

        self.instrument_file_FO.loc[:, "next_expiry"] = self.instrument_file_FO.expiry + pd.DateOffset(days=7)

    @timeit(MetricName="Motilal:add_quantplay_fut_tradingsymbol")
    def add_quantplay_fut_tradingsymbol(self):
        seg_condition = [
            ((self.instrument_file_FO["instrumentname"].str.contains("FUT")) & (
                        self.instrument_file_FO.instrumentname != "OPTFUT"))
        ]

        tradingsymbol = [
            self.instrument_file_FO.tradingsymbol + self.instrument_file_FO.expiry_year + self.instrument_file_FO.month + "FUT"
        ]

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.select(
            seg_condition, tradingsymbol, default=self.instrument_file_FO.tradingsymbol
        )

    @timeit(MetricName="Motilal:add_quantplay_opt_tradingsymbol")
    def add_quantplay_opt_tradingsymbol(self):
        seg_condition = (self.instrument_file_FO["strikeprice"] > 0)
        weekly_option_condition = (
                (self.instrument_file_FO.expiry.dt.month == self.instrument_file_FO.next_expiry.dt.month) & (
            self.instrument_file_FO.exchangename.isin(["NFO", "NSEFO", "NSECD"])))
        month_option_condition = (
                (self.instrument_file_FO.expiry.dt.month != self.instrument_file_FO.next_expiry.dt.month) | (
                    self.instrument_file_FO.exchangename == "MCX"))

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_file_FO.tradingsymbol + self.instrument_file_FO.expiry_year,
            self.instrument_file_FO.tradingsymbol
        )

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.where(
            seg_condition & weekly_option_condition,
            self.instrument_file_FO.tradingsymbol + self.instrument_file_FO.week_option_prefix,
            self.instrument_file_FO.tradingsymbol
        )

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.where(
            seg_condition & month_option_condition,
            self.instrument_file_FO.tradingsymbol + self.instrument_file_FO.month,
            self.instrument_file_FO.tradingsymbol
        )

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_file_FO.tradingsymbol +
            self.instrument_file_FO.strikeprice.astype(float).astype(str).str.split(".").str[0],
            self.instrument_file_FO.tradingsymbol
        )

        self.instrument_file_FO.loc[:, 'after_point'] = (self.instrument_file_FO.strikeprice * 100).astype(int)
        self.instrument_file_FO.loc[:, 'after_point'] = (self.instrument_file_FO.after_point % 100).astype(str).apply(
            lambda x: x.rstrip('0'))
        self.instrument_file_FO.loc[:, 'tradingsymbol'] = np.where(self.instrument_file_FO.after_point.str.len() > 0,
                                                                   self.instrument_file_FO.tradingsymbol + "." + self.instrument_file_FO.after_point,
                                                                   self.instrument_file_FO.tradingsymbol)

        self.instrument_file_FO.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_file_FO.tradingsymbol + self.instrument_file_FO.optiontype,
            self.instrument_file_FO.tradingsymbol
        )

    @timeit(MetricName="Motilal:load_file_by_url")
    def load_file_by_url(self, exchange):
        data_url = "https://openapi.motilaloswal.com/getscripmastercsv?name={}".format(exchange)
        df = pd.read_csv(data_url)
        self.instrument_data_by_exchange[exchange] = df

    @timeit(MetricName="Motilal:load_instrument")
    def load_instrument(self):
        exchanges = ["NSEFO", "MCX", "NSECD", "NSE"]
        threads = []
        for exchange in exchanges:
            th = threading.Thread(target=self.load_file_by_url, args=(exchange,))
            th.start()
            threads.append(th)
        for thread in threads:
            thread.join()

        instrument_file_FO = self.instrument_data_by_exchange["NSEFO"]
        instrument_file_MCX = self.instrument_data_by_exchange["MCX"]
        instrument_file_CDS = self.instrument_data_by_exchange["NSECD"]
        instrument_file_CDS = instrument_file_CDS[instrument_file_CDS.instrumentname.isin(["OPTCUR", "FUTCUR"])]

        instrument_file_FO = pd.concat([instrument_file_CDS, instrument_file_MCX, instrument_file_FO])

        instrument_file_EQ = self.instrument_data_by_exchange["NSE"]
        instrument_file_EQ = instrument_file_EQ[instrument_file_EQ['scripname'].str.contains(" EQ")]

        self.symbol_scripcode_map = pd.Series(instrument_file_EQ.scripcode.values,
                                              index=instrument_file_EQ.scripshortname).to_dict()
        temp = pd.Series(instrument_file_EQ.scripcode.values,
                         index=instrument_file_EQ.scripname).to_dict()
        self.symbol_scripcode_map.update(temp)
        temp = pd.Series(instrument_file_FO.scripcode.values, index=instrument_file_FO.scripname).to_dict()
        self.symbol_scripcode_map.update(temp)

        self.instrument_file_FO = instrument_file_FO
        self.instrument_data = instrument_file_FO

        self.initialize_expiry_fields()
        self.add_quantplay_opt_tradingsymbol()
        self.add_quantplay_fut_tradingsymbol()

        self.quantplay_symbol_map = pd.Series(self.instrument_file_FO.scripname.values,
                                              index=self.instrument_file_FO.tradingsymbol).to_dict()
        self.symbol_to_lot_size_map = pd.Series(self.instrument_file_FO.marketlot.values,
                                                index=self.instrument_file_FO.scripname).to_dict()
        temp = pd.Series(instrument_file_EQ.scripshortname.values,
                         index=instrument_file_EQ.scripshortname).to_dict()
        self.quantplay_symbol_map.update(temp)
        self.initialize_symbol_data()

    @timeit(MetricName="Motilal:initialize_symbol_data")
    def initialize_symbol_data(self):
        instruments = self.instrument_data
        instruments = instruments.to_dict('records')
        self.symbol_data = {}
        for instrument_data in instruments:
            exchange = instrument_data['exchangename']
            tradingsymbol = instrument_data['scripname']

            instrument_data['lot_size'] = instrument_data['marketlot']
            self.symbol_data["{}:{}".format(exchange, tradingsymbol)] = instrument_data

    def get_symbol(self, symbol):
        if symbol not in self.quantplay_symbol_map:
            raise InvalidArgumentException("Symbol {} not found".format(symbol))
        return self.quantplay_symbol_map[symbol]

    def get_exchange(self, exchange):
        if exchange in self.exchange_code_map:
            return self.exchange_code_map[exchange]

        return exchange

    def get_product(self, product):
        return "NORMAL"

    def place_order_quantity(self, quantity, tradingsymbol, exchange):
        lot_size = self.get_lot_size(exchange, tradingsymbol)
        quantity_in_lots = int(quantity / lot_size)

        return quantity_in_lots

    def validate_config(self):
        Constants.logger.info("Validating config file")
        quantplay_config = QplayConfig.get_config()

        if quantplay_config is None:
            return False
        if Motilal.api_key not in quantplay_config['DEFAULT']:
            return False
        if Motilal.two_factor_authentication not in quantplay_config["DEFAULT"]:
            return False
        if Motilal.user_id not in quantplay_config["DEFAULT"]:
            return False
        if Motilal.auth_token not in quantplay_config["DEFAULT"]:
            return False
        if Motilal.secret_key not in quantplay_config["DEFAULT"]:
            return False

        Constants.logger.info("config validation successful")
        return True

    def generate_token(self, user_id, password, api_key, two_fa, totp):
        current_totp = pyotp.TOTP(totp).now()
        Constants.logger.info("TOTP is {}".format(current_totp))
        # initializing string
        str = "{}{}".format(password, api_key)
        result = hashlib.sha256(str.encode())

        data = {
            "userid": user_id,
            "password": result.hexdigest(),
            "2FA": two_fa,
            "totp": current_totp
        }

        self.headers['ApiKey'] = api_key
        self.headers['vendorinfo'] = user_id
        response = requests.post(self.url, headers=self.headers, data=json.dumps(data))

        resp_json = response.json()
        Constants.logger.info("login response {}".format(resp_json))
        self.headers['Authorization'] = resp_json['AuthToken']
        self.user_id = user_id

    def send_otp(self):
        response = requests.post(self.otp_url, headers=self.headers).json()
        Constants.logger.info(response)
        return response

    def verify_otp(self, otp):
        data = {
            "otp": otp
        }
        response = requests.post(self.verify_otp_url, headers=self.headers, data=json.dumps(data)).json()
        Constants.logger.info(response)
        return response

    def get_ltp(self, exchange=None, tradingsymbol=None):
        data = {
            "userid": self.user_id,
            "exchange": exchange,
            "scripcode": self.symbol_scripcode_map[tradingsymbol]
        }

        response = requests.post(self.ltp_utl, headers=self.headers, data=json.dumps(data))
        Constants.logger.info("[GET_LTP_RESPONSE] response {}".format(response.json()))
        return response.json()["data"]["ltp"] / 100.0

    def get_orders(self, order_status=None, order_type=None):
        response = (requests.post(self.order_book_url, headers=self.headers)).json()
        if response["status"] == "ERROR":
            Constants.logger.info("Error while fetching order book [{}]".format(response["message"]))
            raise Exception(response["message"])
        orders = response["data"]

        if order_status:
            orders = [a for a in orders if a['orderstatus'] == order_status]

        if order_type:
            orders = [a for a in orders if a['ordertype'] == order_type]

        return orders

    def get_positions(self):
        response = (requests.post(self.positions_url, headers=self.headers)).json()
        if response["status"] == "ERROR":
            Constants.logger.info("Error while fetching order book [{}]".format(response["message"]))
            raise Exception(response["message"])

        positions = response["data"]

        return positions

    def square_off_all(self, dry_run=True, strike=None, sleep_time=0.1, tag=None):
        positions = self.get_positions()
        if positions == None:
            print("Positions are already squared off")
            return

        positions = pd.DataFrame(positions)
        if strike:
            positions = positions[positions.symbol.str.contains(strike)]

        positions.loc[:, 'net_quantity'] = positions.buyquantity - positions.sellquantity
        positions = positions[['exchange', 'symbol', 'net_quantity']].sort_values("net_quantity")

        positions = positions.to_dict('records')
        orders_to_close = []
        for position in positions:
            exchange = position['exchange']
            symbol = position['symbol']
            quantity = position['net_quantity']

            transaction_type = "SELL"
            if quantity == 0 or exchange != "NSEFO":
                continue
            elif quantity < 0:
                transaction_type = "BUY"

            quantity = abs(quantity)
            quantity_in_lots = int(quantity / self.get_lot_size(exchange, symbol))

            split_into = int(math.ceil(quantity_in_lots / 25))
            split_array = NumberUtils.split(abs(quantity_in_lots), abs(split_into))

            for q in split_array:
                orders_to_close.append(
                    {
                        'symbol': symbol,
                        'exchange': exchange,
                        'transaction_type': transaction_type,
                        'quantity_in_lots': q
                    })

        random.shuffle(orders_to_close)
        orders_to_close = sorted(orders_to_close, key=lambda d: d['transaction_type'])
        for order in orders_to_close:
            print(order['symbol'], order['exchange'], order['transaction_type'], order['quantity_in_lots'])
            if dry_run == False:
                self.place_order(tradingsymbol=order['symbol'],
                                 exchange=order['exchange'],
                                 quantity=order['quantity_in_lots'],
                                 order_type="MARKET",
                                 transaction_type=order['transaction_type'],
                                 tag="square-off",
                                 product="NORMAL",
                                 price=0)
                time.sleep(sleep_time)

    def modify_price(self, order_id, price, trigger_price=None, order_type=None):
        orders = pd.DataFrame(self.get_orders())
        orders = orders.to_dict('records')
        order_found = False

        for order in orders:
            if order['uniqueorderid'] == order_id:
                order_found = True
                break

        if order_found == False:
            Constants.logger.error("[ORDER_NOT_FOUND] invalid modify request for {}".format(order_id))
            return

        order['price'] = price
        if trigger_price != None:
            order['triggerprice'] = trigger_price

        if order["ordertype"] == "Stop Loss":
            order["ordertype"] = "STOPLOSS"

        self.modify_order(order)

    def modify_orders_till_complete(self, orders_placed):
        modification_count = {}
        while 1:
            time.sleep(10)
            orders = pd.DataFrame(self.get_orders())
            orders = orders[orders.uniqueorderid.isin(orders_placed)]

            orders = orders[~orders.orderstatus.isin(["Error", "Traded", "Cancel", "Rejected"])]

            if len(orders) == 0:
                Constants.logger.info("ALL orders have be completed")
                break

            orders = orders.to_dict('records')
            for order in orders:
                order_id = order['uniqueorderid']

                ltp = self.get_ltp(order['exchange'], order['symbol'])
                order['price'] = ltp
                self.modify_order(order)

                if order_id not in modification_count:
                    modification_count[order_id] = 1
                else:
                    modification_count[order_id] += 1

                time.sleep(.1)

                if modification_count[order_id] > 5:
                    order['ordertype'] = "MARKET"
                    order['price'] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)

    def exit_all_trigger_orders(self, tag="ALL",
                                symbol_cotains=None):

        stoploss_orders = self.get_orders(order_status="Confirm", order_type="Stop Loss")

        if len(stoploss_orders) == 0:
            print("All stoploss orders have been already closed")
            return

        stoploss_orders = pd.DataFrame(stoploss_orders)
        if tag != "ALL":
            stoploss_orders = stoploss_orders[stoploss_orders.tag == tag]

        if symbol_cotains is not None:
            stoploss_orders = stoploss_orders[stoploss_orders['symbol'].str.contains(symbol_cotains)]

        if len(stoploss_orders) == 0:
            print("All stoploss orders have been already closed")
            return

        orders_to_close = list(stoploss_orders.uniqueorderid.unique())

        stoploss_orders = stoploss_orders.to_dict('records')
        for stoploss_order in stoploss_orders:
            exchange = stoploss_order['exchange']
            tradingsymbol = stoploss_order['symbol']

            if exchange == "NFO":
                stoploss_order['ordertype'] = "MARKET"
                stoploss_order['price'] = 0
            else:
                ltp = self.get_ltp(exchange, tradingsymbol)
                stoploss_order['ordertype'] = "LIMIT"
                stoploss_order['price'] = self.round_to_tick(ltp)

            self.modify_order(stoploss_order)
            time.sleep(.1)

        self.modify_orders_till_complete(orders_to_close)
        print("All order have been closed successfully")

    def modify_order(self, order):
        data = {
            "uniqueorderid": order['uniqueorderid'],
            "newordertype": order['ordertype'].upper(),
            "neworderduration": order['orderduration'].upper(),
            "newquantityinlot": int(order['totalqtyremaining'] / order['lotsize']),
            # "newdisclosedquantity": 0,
            "newprice": order['price'],
            "newtriggerprice": order['triggerprice'],
            "qtytradedtoday": order['qtytradedtoday'],
            "lastmodifiedtime": order['lastmodifiedtime']
        }

        if 'exchange' in order and order['exchange'] == "MCX":
            data['newquantityinlot'] = int(order['totalqtyremaining'])

        try:
            Constants.logger.info("[MODIFYING_ORDER] order [{}]".format(data))
            response = requests.post(self.modify_order_url, headers=self.headers, data=json.dumps(data)).json()
            Constants.logger.info("[MODIFY_ORDER_RESPONSE] {}".format(response))
        except Exception as e:
            exception_message = "[ORDER_MODIFICATION_FAILED] for {} failed with exception {}".format(
                order['uniqueorderid'],
                e)
            Constants.logger.error("{}".format(exception_message))

    def cancel_order(self, unique_order_id):
        data = {
            "uniqueorderid": unique_order_id
        }

        try:
            Constants.logger.info("Cancelling order [{}]".format(unique_order_id))
            response = requests.post(self.cancel_order_url, headers=self.headers, data=json.dumps(data)).json()
            Constants.logger.info("Cancel order response [{}]".format(response))
        except Exception as e:
            exception_message = "[ORDER_CANCELLATION_FAILED] unique_order_id {} exception {}".format(unique_order_id,
                                                                                                     e)
            Constants.logger.error(exception_message)

    def get_profile(self):
        response = requests.post(self.get_profile_url, headers=self.headers,
                                 data=json.dumps({'Clientcode': self.headers['vendorinfo']})).json()
        if response['status'] == "ERROR":
            raise Exception(response['message'])

        return response['data']

    def margin_summary(self):
        response = requests.post(self.margin_summary_url, headers=self.headers,
                                 data=json.dumps({'Clientcode': self.headers['vendorinfo']})).json()
        if response['status'] == "ERROR":
            raise Exception(response['message'])

        return response['data']

    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        data = {
            "exchange": exchange,
            "symboltoken": self.symbol_scripcode_map[tradingsymbol],
            "buyorsell": transaction_type,
            "ordertype": order_type,
            "producttype": product,
            "orderduration": "DAY",
            "price": price,
            "triggerprice": trigger_price,
            "quantityinlot": quantity,
            "disclosedquantity": 0,
            "amoorder": "N",
            "algoid": "",
            "tag": tag
        }
        try:
            Constants.logger.info("[PLACING_ORDER] {}".format(json.dumps(data)))
            response = requests.post(self.place_order_url, headers=self.headers, data=json.dumps(data)).json()
            Constants.logger.info("[PLACE_ORDER_RESPONSE] {} input {}".format(response, json.dumps(data)))
            if response['status'] == "ERROR":
                raise Exception(response['message'])
            return response['uniqueorderid']
        except Exception as e:
            exception_message = "Order placement failed with error [{}]".format(str(e))
            print(exception_message)

    def on_message(self, ws, order):
        try:
            order = json.loads(order)
            print(order)

            order['placed_by'] = self.user_id
            order['tag'] = self.user_id
            order['order_id'] = order['orderid']
            order['exchange_order_id'] = order['order_id']
            order['transaction_type'] = order['transactiontype']
            order['quantity'] = int(order['quantity'])
            order['order_type'] = order['ordertype']

            if order['exchange'] == "NFO":
                order["tradingsymbol"] = self.symbol_map[order["tradingsymbol"]]

            if order['order_type'] == "STOPLOSS_LIMIT":
                order['order_type'] = "SL"

            if 'triggerprice' in order and order['triggerprice'] != 0:
                order['trigger_price'] = float(order['triggerprice'])
            else:
                order['trigger_price'] = None

            if order["status"] == "trigger pending":
                order["status"] = "TRIGGER PENDING"
            elif order["status"] == "cancelled":
                order["status"] = "CANCELLED"
            elif order["status"] == "open":
                order["status"] = "OPEN"
            elif order["status"] == "complete":
                order["status"] = "COMPLETE"

            # self.order_updates.put(order)
            print("Final order {}".format(order))
        except Exception as e:
            Constants.logger.error("[ORDER_UPDATE_PROCESSING_FAILED] {}".format(e))
        print(json.dumps(order))

    def on_error(self, ws, error):
        print(error)

    def on_close(self, ws):
        print("### closed ###")

    def on_open(self, ws):
        def run(*args):
            for i in range(300000):
                time.sleep(1)
                print({
                    "actiontype": "TradeSubscribe",
                    "clientid": self.headers['vendorinfo']
                })
                ws.send(json.dumps({
                    "actiontype": "TradeSubscribe",
                    "clientid": self.headers['vendorinfo']
                }))
            time.sleep(1)
            ws.close()
            print("thread terminating...")

        thread.start_new_thread(run, ())

    def stream_order_data(self):
        api_key = self.headers['ApiKey']
        auth_token = self.headers['Authorization']
        client_id = self.headers['vendorinfo']

        root_url = "wss://openapi.motilaloswal.com/ws"
        ws_url = "{}?clientid={}&authtoken={}&apikey={}".format(root_url, client_id, auth_token, api_key)

        websocket.enableTrace(False)
        print(ws_url)
        ws = websocket.WebSocketApp(ws_url,
                                    on_message=self.on_message,
                                    on_error=self.on_error)
        ws.on_open = self.on_open
        ws.run_forever()

    def account_summary(self):
        response = {
            'margin_used': 0,
            'pnl' : 0
        }

        margin_summary = self.margin_summary()
        for margin_particular in margin_summary:
            if margin_particular['srno'] == 700 or margin_particular['srno'] == 600:
                response['pnl'] += margin_particular['amount']
            if margin_particular['srno'] == 103:
                response['margin_available'] = margin_particular['amount']
            if margin_particular['srno'] in [301, 321, 340, 360]:
                response['margin_used'] += margin_particular['amount']

        return response

    def profile(self):
        api_response = requests.post(self.get_profile_url, headers=self.headers,
                                     data=json.dumps({'Clientcode': self.headers['vendorinfo']})).json()
        if api_response['status'] == "ERROR":
            raise Exception(api_response['message'])

        api_response = api_response['data']
        response = {
            'user_id': api_response['clientcode'],
            'full_name': api_response['name'],
            'segments': api_response['exchanges']
        }

        return response

    def positions(self):
        response = (requests.post(self.positions_url, headers=self.headers)).json()
        if response["status"] == "ERROR":
            Constants.logger.info("Error while fetching order book [{}]".format(response["message"]))
            raise Exception(response["message"])

        positions = response["data"]
        positions = pd.DataFrame(positions)

        return positions

    def orders(self, tag=None):
        response = (requests.post(self.order_book_url, headers=self.headers)).json()
        if response["status"] == "ERROR":
            Constants.logger.info("Error while fetching order book [{}]".format(response["message"]))
            raise Exception(response["message"])
        orders = response["data"]
        orders = pd.DataFrame(orders)

        positions = self.positions()
        if len(positions) == 0:
            return positions

        orders = pd.merge(orders, positions[['symbol', 'LTP']],
                          how="left",
                          left_on=['symbol'],
                          right_on=['symbol'])

        orders.rename(
            columns={
                'clientid': 'user_id',
                "LTP": 'ltp',
                'buyorsell': 'transaction_type',
                "averageprice": "average_price",
                "totalqtytraded": "quantity_traded"
            }, inplace=True)
        orders.loc[:, 'tradingsymbol'] = orders.symbol
        orders.loc[:, 'product'] = orders.ordercategory

        orders = orders[['user_id', 'tradingsymbol', 'tag', 'quantity_traded', 'average_price', 'transaction_type',
                         'ltp', 'exchange', 'product']]
        orders.loc[:, 'pnl'] = orders.ltp * orders.quantity_traded - orders.average_price * orders.quantity_traded
        orders.loc[:, 'pnl'] = np.where(orders.transaction_type == "SELL", -orders.pnl, orders.pnl)

        if tag:
            orders = orders[orders.tag == tag]

        return orders

